"""
@Description: 
@Usage: 
@Author: liuxianglong
@Date: 2022/5/14 下午9:06
"""
import socket
import requests
from .api import API


class BaseInfo(object):

    def __init__(self, crawler, spider):
        ip_type = crawler.settings.get('CS_HOST_IP_TYPE', 'internal')
        if ip_type == 'internal':
            get_ip = self.get_internal_ip
        else:
            get_ip = self.get_external_ip
        self.host = get_ip()
        self.project = crawler.settings.get('BOT_NAME')
        self.spider = spider.name
        self.job_id = getattr(spider, '_job', '000000')
        self.stats = crawler.stats
        self.api = API.from_crawler(crawler)
        self.enable_send_err_text = crawler.settings.getbool('CS_ENABLE_SEND_ERR_TEXT', True)

    @staticmethod
    def get_internal_ip():
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            s.connect(('8.8.8.8', 80))
            ip = s.getsockname()[0]
            return ip

    @staticmethod
    def get_external_ip():
        ip = requests.get('https://api.ipify.org').text
        return ip



"""
@Description: 
@Usage: 
@Author: liuxianglong
@Date: 2022/5/13 上午1:11
"""
import logging
import requests

logger = logging.getLogger(__name__)


def cs_api(func):

    def wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Exception as e:
            logger.warning(f'API {func.__name__} error, {e}')

    return wrapper


class API(object):

    def __init__(self, settings):
        self._settings = settings
        self._host = settings.get('CS_BACKEND', 'http://localhost:8000')
        self._spider_stats_end_point = f"{self._host}/api/v1/scrapyd/spiderStats/"
        self._errlog_rate_end_point = f"{self._host}/api/v1/logs/errorLogRate/"
        self._errlog_content_end_point = f"{self._host}/api/v1/logs/errorLogContent/"
        self._api_token = settings.get('CS_API_TOKEN', '')

    @classmethod
    def from_crawler(cls, crawler):
        o = cls(crawler.settings)
        return o

    @property
    def headers(self):
        return {
            'Content-Type': 'application/json',
            'Authorization': f'Token {self._api_token}'
        }

    @cs_api
    def send_stats_data(self, raw_json):
        response = requests.post(self._spider_stats_end_point, headers=self.headers, json=raw_json)
        logger.info(f'Send scrapy stats success, {response.text}')

    @cs_api
    def send_errlog_rate(self, raw_json):
        response = requests.post(self._errlog_rate_end_point, headers=self.headers, json=raw_json)
        logger.info(f'Send errlog rate success, {response.text}')

    @cs_api
    def send_errlog_content(self, raw_json):
        response = requests.post(self._errlog_content_end_point, headers=self.headers, json=raw_json)
        logger.info(f'Send errlog content success, {response.text}')




"""
@Description: 
@Usage: 
@Author: liuxianglong
@Date: 2022/5/14 上午11:52
"""
import datetime
import logging
from twisted.internet import task
from .core import BaseInfo

logger = logging.getLogger(__name__)


class LogHandler(logging.Handler, BaseInfo):
    """Record log levels count into a crawler stats"""

    def __init__(self, crawler, spider, *args, **kwargs):
        super().__init__(*args, **kwargs)
        BaseInfo.__init__(self, crawler, spider)
        self.now_date = None
        self.now_hour = None

    def emit(self, record):
        today = datetime.datetime.now().date().strftime('%Y-%m-%d')
        if today != self.now_date:
            self.now_date = today
            self.stats.set_value('log_count/log_daily_total_count', 0)
            self.stats.set_value('log_count/log_daily_error_count', 0)

        hour = datetime.datetime.now().hour
        if hour != self.now_hour:
            self.now_hour = hour
            self.stats.set_value('log_count/log_hourly_total_count', 0)
            self.stats.set_value('log_count/log_hourly_error_count', 0)

        self.stats.inc_value('log_count/log_daily_total_count')
        self.stats.inc_value('log_count/log_hourly_total_count')

        if record.levelname in ('ERROR', 'CRITICAL'):
            self.stats.inc_value('log_count/log_daily_error_count')
            self.stats.inc_value('log_count/log_hourly_error_count')

            if self.enable_send_err_text:
                d = {
                    "host": self.host,
                    "project": self.project,
                    "spider": self.spider,
                    "job_id": self.job_id,
                    "record_time": datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                    "content": record.message,
                    "level": record.levelname,
                    "log_time": str(record.asctime),
                    "module": record.name,
                    "lineno": record.lineno,
                    "exc_info": record.exc_info,
                    "func_name": record.funcName,
                }
                self.api.send_errlog_content(raw_json=d)


class ErrorLogRate(BaseInfo):

    def __init__(self, crawler, spider):
        BaseInfo.__init__(self, crawler, spider)
        self.task = None
        self.interval = 60

    def spider_open(self, spider):
        self.task = task.LoopingCall(self.send, spider)
        self.task.start(self.interval)

    def spider_close(self, spider, reason):
        if self.task and self.task.running:
            self.send(spider)
            self.task.stop()

    def send(self, spider):
        today = datetime.datetime.now().date().strftime('%Y-%m-%d')
        hour = datetime.datetime.now().hour

        d = dict()
        d['host'] = self.host
        d['project'] = self.project
        d['spider'] = self.spider
        d['job_id'] = self.job_id
        d['log_date'] = today
        d['record_time'] = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')

        d['log_total_count'] = self.stats.get_value('log_count/log_daily_total_count', 0)
        d['log_error_count'] = self.stats.get_value('log_count/log_daily_error_count', 0)
        if d['log_total_count'] != 0:
            d['log_error_rate'] = round(d['log_error_count']/d['log_total_count'], 4)
        else:
            d['log_error_rate'] = 0
        self.api.send_errlog_rate(raw_json=d)

        d['log_total_count'] = self.stats.get_value('log_count/log_hourly_total_count', 0)
        d['log_error_count'] = self.stats.get_value('log_count/log_hourly_error_count', 0)
        if d['log_total_count'] != 0:
            d['log_error_rate'] = round(d['log_error_count']/d['log_total_count'], 4)
        else:
            d['log_error_rate'] = 0
        d['log_hour'] = hour
        self.api.send_errlog_rate(raw_json=d)




"""
@Description: 
@Usage: 
@Author: liuxianglong
@Date: 2022/5/12 下午10:13
"""
import json
import logging
import datetime
from twisted.internet import task
from .core import BaseInfo

logger = logging.getLogger(__name__)


class SpiderStats(BaseInfo):
    def __init__(self, crawler, spider):
        BaseInfo.__init__(self, crawler, spider)
        self.task = None
        self.interval = 60

    def spider_open(self, spider):
        self.stats.set_value('host', self.host)
        self.stats.set_value('project', self.project)
        self.stats.set_value('spider', self.spider)
        self.stats.set_value('job_id', self.job_id)
        self.task = task.LoopingCall(self.send, spider)
        self.task.start(self.interval)

    def spider_close(self, spider, reason):
        if self.task and self.task.running:
            self.send(spider, finish_reason=reason, close=True)
            self.task.stop()

    def send(self, spider, finish_reason=None, close=False):
        stats = self.stats.get_stats()
        new_stats = {}
        for k, v in stats.items():
            if k == 'start_time' and type(v) != str:
                v = str(stats[k] + datetime.timedelta(hours=8))

            if k == 'finish_time' and type(v) != str:
                v = str(stats[k] + datetime.timedelta(hours=8))

            if k.startswith('memusage') and type(v) != str:
                v = str(round(v/(1024*1024), 1)) + 'MB'

            new_stats[k] = v

        d = dict()
        d['host'] = self.host
        d['project'] = self.project
        d['spider'] = self.spider
        d['job_id'] = self.job_id
        d['record_time'] = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')

        d['log_debug_count'] = new_stats.get('log_count/DEBUG', 0)
        d['log_info_count'] = new_stats.get('log_count/INFO', 0)
        d['log_warn_count'] = new_stats.get('log_count/WARNING', 0)
        d['log_error_count'] = new_stats.get('log_count/ERROR', 0)
        d['log_critical_count'] = new_stats.get('log_count/CRITICAL', 0)

        log_total_count = new_stats.get('log_count/log_hourly_total_count', 0)
        log_error_count = new_stats.get('log_count/log_hourly_error_count', 0)
        if log_total_count != 0:
            d['log_hourly_error_rate'] = round(log_error_count/log_total_count, 4)
        else:
            d['log_hourly_error_rate'] = 0
        d['memory_use'] = float(new_stats.get('memusage/max', '0MB').replace('MB', '').strip())
        d['start_time'] = new_stats.get('start_time')
        d['finish_time'] = new_stats.get('finish_time', None)
        d['finish_reason'] = finish_reason
        d['elapsed_time'] = new_stats.get('elapsed_time_seconds', None)
        d['state'] = 1 if not close else 0
        d['stats'] = json.dumps(new_stats, ensure_ascii=False)
        self.api.send_stats_data(raw_json=d)



"""
@Description: 
@Usage: 
@Author: liuxianglong
@Date: 2022/5/12 下午10:24
"""
from scrapy import signals
import logging
from .stats import SpiderStats
from .log import LogHandler, ErrorLogRate

logger = logging.getLogger(__name__)


class ScrapyMonitor(object):

    def __init__(self, crawler):
        self.crawler = crawler

    @classmethod
    def from_crawler(cls, crawler):
        o = cls(crawler)

        # connect the oension object to signals
        crawler.signals.connect(o.spider_opened,
                                signal=signals.spider_opened)

        crawler.signals.connect(o.spider_closed,
                                signal=signals.spider_closed)

        crawler.signals.connect(o.log_request,
                                signal=signals.request_reached_downloader)

        crawler.signals.connect(o.log_response,
                                signal=signals.response_downloaded)

        crawler.signals.connect(o.item_scraped,
                                signal=signals.item_scraped)

        crawler.signals.connect(o.item_dropped,
                                signal=signals.item_dropped)

        crawler.signals.connect(o.item_error,
                                signal=signals.item_error)

        return o

    def spider_opened(self, spider):
        self.stats_sender = SpiderStats(self.crawler, spider)
        self.stats_sender.spider_open(spider)

        self.errlog_rate_sender = ErrorLogRate(self.crawler, spider)
        self.errlog_rate_sender.spider_open(spider)

        handler = LogHandler(self.crawler, spider, level=self.crawler.settings.get('LOG_LEVEL'))
        logging.root.addHandler(handler)

    def spider_closed(self, spider, reason):
        self.stats_sender.spider_close(spider, reason)
        self.errlog_rate_sender.spider_close(spider, reason)

    def log_request(self, request, spider):
        pass

    def log_response(self, response, request, spider):
        pass

    def item_scraped(self, item, response, spider):
        pass

    def item_dropped(self, item, response, spider):
        pass

    def item_error(self, item, response, spider):
        pass
